from pam import client


def create_and_get_database(arango_conn, database_name, username, password):
    """Returns database object. If not exists, create a new one. Requires username & passwords
    
    :parameters:
    - `database_name`: database name to create (string)
    - `username`: username with database creation role (string)
    - `password`: password with database creation role (string)
    """
    
    sys_db = arango_conn.db('_system', username=username, password=password)

    if not sys_db.has_database(database_name):
        sys_db.create_database(database_name)
    
    return arango_conn.db(database_name, username=username, password=password)


import itertools
from pam import aql, database


def dict_spliter(target_dict, num_split=1000):
    """Split dictionary into chunks and return as list
    
    :parameters:
    - `target_dict`: dictionary to split (dict)
    - `num_split` (optional): chunk size to split (integer)
    """
    
    result = []

    for i in range(len(target_dict)//num_split +1):
        result.append(itertools.islice(target_dict.items(), i * num_split, (i+1) * num_split))

    return result


def arango_split_task(database_obj, partial_query, params, num_split=1000):
    """Split AQLs into chunks in order to adjust performance. 1000 worked best for me
    It is a system function used in converter. Users may not use this directly.
    
    :parameters:
    - `database_obj`: database object in which to execute aql (Arango Database Object)
    - `partial_query`: partial AQL to split (string)
    - `params`: params to use (dict)
    - `num_split` (optional): chunk size to split (integer)
    """

    for partial_rows in dict_spliter(params['rows'], num_split):
        copied = params.copy()
        copied['rows'] = [i for i in dict(partial_rows).values()]
        final_query = partial_query.format(**copied)

        aql.execute_aql(final_query, database_obj, show_query=True)



import logging
from arango import ArangoClient
from arango.http import HTTPClient
from arango.response import Response
from pam import database 
from requests import Session
from requests.packages.urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter


class CustomHTTPClient(HTTPClient):
    def __init__(self):
        self._logger = logging.getLogger('my_logger')

    def create_session(self, host):
        session = Session()

        # Add request header.
        session.headers.update({'x-my-header': 'true'})

        # Enable retries.
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            method_whitelist=["HEAD", "GET", "OPTIONS"],
        )
        http_adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount('https://', http_adapter)
        session.mount('http://', http_adapter)

        return session

    def send_request(self,
                     session,
                     method,
                     url,
                     params=None,
                     data=None,
                     headers=None,
                     auth=None):
        # Add your own debug statement.
        self._logger.debug(f'Sending request to {url}')

        # Send a request.
        response = session.request(
            method=method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            auth=auth,
            timeout=600
        )
        self._logger.debug(f'Got {response.status_code}')

        # Return an instance of arango.response.Response.
        return Response(
            method=response.request.method,
            url=response.url,
            headers=response.headers,
            status_code=response.status_code,
            status_text=response.reason,
            raw_body=response.text,
        )


def get_arango_conn(hosts):
    """Returns ArangoDB Client Connection Object
    
    :parameters:
    - `hosts`: host of ArangoDB (string)
    """

    arango_conn = ArangoClient(
        hosts=hosts,
        http_client=CustomHTTPClient()
    )

    return arango_conn




import logging
from pam import database

def execute_aql(query, database_obj, show_query=False, bind_vars=None):
    """Execute AQL with bind_vars and return ArangoDB cursor
    
    :parameters:
    - `query`: AQL to Execute (string)
    - `database_obj`: database object in which to execute aql (Arango Database Object)
    - `show_query` (optional): whether to print full aql (boolean)
    - `bind_vars` (optional): bind variables to supply to AQL (dict)
    """

    temp = query
    if bind_vars:
        for var in bind_vars:
            if type(bind_vars[var]) == list:
                temp = temp.replace('@{}'.format(var), str(bind_vars[var]))
            else:
                temp = temp.replace('@{}'.format(var), "'{}'".format(str(bind_vars[var])))

    cursor = database_obj.aql.execute(query, bind_vars=bind_vars)

    if show_query:
        logging.warning(temp)

    logging.warning(str(cursor.statistics()))

    return cursor


import logging
import time
import uuid

from pam import collection
from pam import utils


def arango_converter(list_of_dict_data, database_obj, schemas, mapping_list):
    """ 
    """

    collections = {}
    
    #1. Loop mapping_list and initiate target collections & indices
    for mapping in mapping_list:
        coll_def = schemas[mapping]
        coll_name = coll_def['collection']
        _type, _type2 = coll_def['type']
        
        coll_obj = collection.create_and_get_collection(database_obj, coll_name, True if _type == 'edge' else False)

        collections[coll_name] = {}

        # Configuration for conditions
        if coll_def.get('condition'):
            cond = coll_def['condition']

            if cond.get('min_by'):
                for min_by_field in cond.get('min_by'):
                    coll_def['index'].append({'field': ('_key', min_by_field), 'unique': False, 'ttl': False})

            if cond.get('max_by'):
                for max_by_field in cond.get('max_by'):
                    coll_def['index'].append({'field': ('_key', max_by_field), 'unique': False, 'ttl': False})

            if cond.get('min_by') and coll_def.get('max_by'):
                coll_def['index'].append({'field': ('_key', min_by_field, max_by_field,), 'unique': False, 'ttl': False})
        
        # Index Initialization
        for index in coll_def['index']:
            if index.get('ttl'):
                collection.add_ttl_index(coll_obj, index['field'], index['unique'], index['ttl'])
            else:
                collection.add_persistent_index(coll_obj, index['field'], index['unique'])

    step_2 = time.time()

    #2. 프레스토 커서를 순회하면서 스키마별로 업데이트할 도큐먼트를 분류함. 

    step_3 = time.time()
    for idx, row in enumerate(list_of_dict_data):
        if idx == 0:
            step_3 = time.time()
            logging.warning("...presto execution time : {} secs".format(str(round(step_3 - step_2))))
        
        print(row)
        for mapping in mapping_list:
            coll_def = schemas[mapping]

            doc = {}

            _type, _type2 = coll_def['type']

            ### _key, _from, _to Init
            if _type2 == 'unique_vertex':
                doc['_key'] = '_'.join([str(row[i]) for i in coll_def['unique_key']])
                if doc['_key'] == 'None':
                    continue

            elif _type2 == 'unique_edge_on_event':
                for field in coll_def['_from']:
                    if not row.get(field):
                        continue
                for field in coll_def['_to']:
                    if not row.get(field):
                        continue

                doc['_from'] = coll_def['_from_collection'] + '/' + '_'.join([str(row[i]) for i in coll_def['_from']])
                doc['_to'] = coll_def['_to_collection'] + '/' + '_'.join([str(row[i]) for i in coll_def['_to']])
                doc['_key'] = '_'.join([str(row[i]) for i in coll_def['_from']]) + '_' + '_'.join([str(row[i]) for i in coll_def['_to']]) + '_' + '_'.join([str(row[i]) for i in coll_def['unique_key']])

            elif _type2 == 'unique_edge_btw_vertices':
                for field in coll_def['_from']:
                    if not row.get(field):
                        continue
                for field in coll_def['_to']:
                    if not row.get(field):
                        continue
                doc['_from'] = coll_def['_from_collection'] + '/' + '_'.join([str(row[i]) for i in coll_def['_from']])
                doc['_to'] = coll_def['_to_collection'] + '/' + '_'.join([str(row[i]) for i in coll_def['_to']])
                doc['_key'] = '_'.join([str(row[i]) for i in coll_def['_from']]) + '_' + '_'.join([str(row[i]) for i in coll_def['_to']])

            elif _type2 == 'unique_edge_from_vertex':
                for field in coll_def['_from']:
                    if not row.get(field):
                        continue
                for field in coll_def['_to']:
                    if not row.get(field):
                        continue
                doc['_key'] = '_'.join([str(row[i]) for i in coll_def['_from']])
                doc['_from'] = coll_def['_from_collection'] + '/' + '_'.join([str(row[i]) for i in coll_def['_from']])
                doc['_to'] = coll_def['_to_collection'] + '/' + '_'.join([str(row[i]) for i in coll_def['_to']])

            if doc['_key'] is None:
                continue
            
            for k in ['_key', '_from', '_to']:
                if doc.get(k):
                    doc[k] = doc[k].replace(' ', '_')

            if coll_def.get('condition'):
                doc['unique_identifier'] = doc['_key'] + str(uuid.uuid4())
            else:
                doc['unique_identifier'] = doc['_key']
            

            ### fields Init
            for field in coll_def['fields']:
                if row.get(coll_def['fields'][field]):
                    doc[field] = row[coll_def['fields'][field]]

            ### Add Doc to Collection
            collections[coll_def['collection']][doc['unique_identifier']] = doc

    step_4 = time.time()
    print("...cursor iter time : {} secs".format(str(round(step_4 - step_3))))

    #3. 각 스키마의 타입에 따라 기본 AQL을 세팅하고, CRUD 작업 실행
    for mapping in mapping_list:
        coll_def = schemas[mapping]

        base_query = ""

        if coll_def.get('condition'):
            cond = coll_def['condition']
            if cond.get('min_by'):
                for min_by_field in cond.get('min_by'):
                    base_query += """
                    '{min_by_field}' : doc.{min_by_field} < OLD.{min_by_field} ? doc.{min_by_field} : OLD.{min_by_field}""".format(min_by_field=min_by_field)
                    
                    for field in cond.get('min_by').get(min_by_field):
                        base_query += """,
                        '{field}' : doc.{min_by_field} < OLD.{min_by_field} ? doc.{field} : OLD.{field}
                        """.format(field=field, min_by_field=min_by_field)

            if cond.get('max_by'):
                for max_by_field in cond.get('max_by'):
                    base_query += """
                    '{max_by_field}' : doc.{max_by_field} > OLD.{max_by_field} ? doc.{max_by_field} : OLD.{max_by_field}""".format(max_by_field=max_by_field)
                    
                    for field in cond.get('max_by').get(max_by_field):
                        base_query += """,
                        '{field}' : doc.{max_by_field} > OLD.{max_by_field} ? doc.{field} : OLD.{field}
                        """.format(field=field, max_by_field=max_by_field)

            if cond.get('if'):
                for if_clause in cond.get('if'):
                    if base_query:
                        base_query += ","
                    base_query += if_clause

        if base_query:
            op = """
            UPDATE {{
                {}
            }}
            """.format(base_query)

            target_aql = """
            FOR doc in {rows}

            UPSERT {{_key : doc._key}}
                INSERT doc
                {op}

            IN {collection} 

            OPTIONS {{
                exclusive : true
            }}"""

        else:
            op = 'INSERT doc'

            target_aql = """
            FOR doc in {rows}

            {op}

            IN {collection} 

            OPTIONS {{
                overwriteMode : "ignore",
                exclusive : true
            }}
            """
        if collections.get(coll_def['collection']):
            # unique_identifier 필드 제거
            [v.pop('unique_identifier') for k, v in collections[coll_def['collection']].items()]

            params = {
                'rows' : collections[coll_def['collection']],
                'op' : op,
                'collection' : coll_def['collection']
            }
            #celery 통해 작업 분할
            utils.arango_split_task(database_obj, target_aql, params)
            
            del collections[coll_def['collection']]
    
    step_5 = time.time()
    print("...Arango Op time : {} secs".format(str(round(step_5 - step_4, 3))))


from setuptools import find_packages, setup
import setuptools

install_requires = [
    'bleach==4.1.0',
    'certifi==2021.10.8',
    'cffi==1.15.0',
    'charset-normalizer==2.0.9',
    'colorama==0.4.4',
    'cryptography==36.0.1',
    'dataclasses==0.8',
    'docutils==0.18.1',
    'idna==3.3',
    'importlib-metadata==4.8.3',
    'jeepney==0.7.1',
    'keyring==23.4.0',
    'packaging==21.3',
    'pkg_resources==0.0.0',
    'pkginfo==1.8.2',
    'pycparser==2.21',
    'Pygments==2.10.0',
    'PyJWT==2.3.0',
    'pyparsing==3.0.6',
    'python-arango==7.3.0',
    'readme-renderer==32.0',
    'requests==2.26.0',
    'requests-toolbelt==0.9.1',
    'rfc3986==1.5.0',
    'SecretStorage==3.3.1',
    'setuptools-scm==6.3.2',
    'six==1.16.0',
    'tomli==1.2.3',
    'tqdm==4.62.3',
    'twine==3.7.1',
    'typing_extensions==4.0.1',
    'urllib3==1.26.7',
    'webencodings==0.5.1',
    'zipp==3.6.0'
]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='python-arango-mapper',
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    install_requires=install_requires,
    version='0.1.3',
    description='fast and easy-to-use python-arango mapper library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='lee ui dam',
    author_email='ud803da@gmail.com',
    url='https://github.com/ud803/python-arango-mapper',
    license='GNU General Public License v3.0'
)



from pam import database


def create_and_get_collection(database_obj, collection_name, edge):
    """Returns collection object. If not exists, create a new one. 
    
    :parameters:
    - `database_obj`: database object in which to execute aql (Arango Database Object)
    - `collection_name`: collection name to create (string)
    - `edge` : which collection type to create, if True, makes edge, else makes vertex (boolean)
    """
        
    if not database_obj.has_collection(collection_name):
        database_obj.create_collection(collection_name, edge=edge)
    
    return database_obj.collection(collection_name)

    
def add_persistent_index(coll_obj, fields, unique, in_background=True):
    """Add persistent index to collection
    
    :parameters:
    - `coll_obj`: collection object in which to add index (Arango Collection Object)
    - `fields`: list of fields to make index (list of strings)
    - `unique`: whether field sets are unique (boolean)
    - `in_background`: whether to create index in background or not (boolean)
    """

    coll_obj.add_persistent_index(fields=fields, unique=unique, in_background=in_background)

    return


def add_ttl_index(coll_obj, field, unique, seconds, in_background=True):
    """Add ttl index to collection. 
    
    :parameters:
    - `coll_obj`: collection object in which to add index (Arango Collection Object)
    - `field`: ttl field name to create. raw data in ttl field should be datetime object (string)
    - `unique`: whether field sets are unique (boolean)
    - `seconds`: seconds until document expires (integer)
    - `in_background`: whether to create index in background or not (boolean)
    """

    coll_obj.add_ttl_index(fields=[field], expiry_time=seconds, unique=unique, in_background=in_background)

    return
